@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2000-2004  Ulric Eriksson <ulric@siag.nu>
+   Copyright (C) 2000-2006  Ulric Eriksson <ulric@siag.nu>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -42,6 +42,9 @@
 #ifdef HAVE_POLL
 #include <sys/poll.h>
 #endif	/* HAVE_POLL */
+#ifdef HAVE_KQUEUE
+#include <sys/event.h>
+#endif
 #include <fcntl.h>
 #include <unistd.h>
 #include <syslog.h>
@@ -84,6 +87,9 @@ typedef struct {
 	int downn, upn;
 	int clt;
 	int index;		/* server index */
+#ifdef HAVE_KQUEUE
+	int i;			/* connection index */
+#endif
 #ifdef HAVE_SSL
 	SSL *ssl;
 #endif
@@ -122,6 +128,7 @@ static server *servers;
 static acl *acls[ACLS_MAX];
 static int emerg_server = -1;	/* server of last resort */
 static int emergency = 0;	/* are we using the emergency server? */
+static int abuse_server = -1;	/* server for naughty clients */
 static connection *conns;
 
 static int debuglevel;
@@ -148,6 +155,7 @@ static int delayed_forward = 0;
 static int do_stats = 0;
 static int do_restart_log = 0;
 static int use_poll = 0;
+static int use_kqueue = 0;
 static int http = 0;
 static int client_acl, control_acl;
 
@@ -165,6 +173,7 @@ static char *listenport = NULL;
 static char *ctrlport = NULL;
 static int listenfd, ctrlfd;
 static char *e_server = NULL;
+static char *a_server = NULL;
 static char *jail = NULL;
 static char *user = NULL;
 
@@ -1047,6 +1056,13 @@ static void store_conn(int downfd, int upfd, int clt, int index)
 #endif
 		conns[i].clt = clt;
 		conns[i].index = index;
+#ifdef HAVE_KQUEUE
+		/* kqueue can store a pointer to arbitrary data with each
+		 * event, but it must be a pointer. We store a pointer to
+		 * conns[i]. copy_up/down then require i again, which we
+		 * can't get without storing i within the struct itself. */
+		conns[i].i = i;
+#endif
 		current = index;
 	} else {
 		if (debuglevel)
@@ -1092,11 +1108,13 @@ static void usage(void)
 	printf("usage:\n"
 	       "  pen [-C addr:port] [-X] [-b sec] [-S N] [-c N] [-e host[:port]] \\\n"
 	       "	  [-t sec] [-x N] [-w dir] [-HPWadfhrs] \\\n"
+	       "          [-o option] \\\n"
 	       "	  [-E certfile] [-K keyfile] \\\n"
 	       "	  [-G cacertfile] [-A cacertdir] \\\n"
 	       "	  [-Z] [-R] [-L protocol] \\\n"
 	       "	  [host:]port h1[:p1[:maxc1[:hard1]]] [h2[:p2[:maxc2[:hard2]]]] ...\n"
 	       "\n"
+	       "  -B host:port abuse server for naughty clients\n"
 	       "  -C port   control port\n"
 	       "  -T sec    tracking time in seconds (0 = forever) [%d]\n"
 	       "  -H	add X-Forwarded-For header in http requests\n"
@@ -1108,7 +1126,7 @@ static void usage(void)
 	       "  -S N      max number of servers [%d]\n"
 	       "  -c N      max number of clients [%d]\n"
 	       "  -d	debugging on (repeat -d for more)\n"
-	       "  -e host:port emergency server of last resourt\n"
+	       "  -e host:port emergency server of last resort\n"
 	       "  -f	stay in foregound\n"
 	       "  -h	use hash for initial server selection\n"
 	       "  -j dir    run in chroot\n"
@@ -1122,6 +1140,7 @@ static void usage(void)
 	       "  -p file   write pid to file\n"
 	       "  -x N      max number of simultaneous connections [%d]\n"
 	       "  -w file   save statistics in HTML format in a file\n"
+	       "  -o option use option in penctl format\n"
 	       "  -E certfile   use the given certificate in PEM format\n"
 	       "  -K keyfile    use the given key in PEM format (may be contained in cert)\n"
 	       "  -G cacertfile file containing the CA's certificate\n"
@@ -1164,7 +1183,8 @@ static void init(int argc, char **argv)
 	conns = pen_calloc(connections_max, sizeof *conns);
 	clients = pen_calloc(clients_max, sizeof *clients);
 	/* one extra server slot for the emergency server */
-	servers = pen_calloc(servers_max+1, sizeof *servers);
+	/* and one for the abuse server */
+	servers = pen_calloc(servers_max+2, sizeof *servers);
 
 	nservers = 0;
 	current = 0;
@@ -1207,6 +1227,18 @@ static void init(int argc, char **argv)
 		server++;
 	}
 
+	if (a_server) {
+		abuse_server = server;
+		servers[server].status = 0;
+		servers[server].c = 0;	/* connections... */
+		setaddress(&servers[server].addr, &servers[server].port,
+			   a_server, port,
+			   &servers[server].maxc, &servers[server].hard);
+		servers[server].sx = 0;
+		servers[server].rx = 0;
+		server++;
+	}
+
 	for (i = 0; i < clients_max; i++) {
 		clients[i].last = 0;
 		clients[i].addr.s_addr = 0;
@@ -1263,7 +1295,7 @@ static int try_server(int index, int sticky, unsigned int cli_addr)
 		return -1;
 	}
 	upfd = socket(AF_INET, SOCK_STREAM, 0);
-	if (upfd < 0) error("Error opening socket");
+	if (upfd < 0) error("Error opening socket: %d", errno);
 	memset(&serv_addr, 0, sizeof serv_addr);
 	serv_addr.sin_family = AF_INET;
 	serv_addr.sin_addr.s_addr = servers[index].addr.s_addr;
@@ -1342,6 +1374,7 @@ static void write_cfg(char *p)
 	if (clients_max != CLIENTS_MAX)
 		fprintf(fp, " -c %d", clients_max);
 	if (e_server) fprintf(fp, " -e %s", e_server);
+	if (a_server) fprintf(fp, " -B %s", a_server);
 	if (jail) fprintf(fp, " -j '%s'", jail);
 	if (pidfile) fprintf(fp, " -p '%s'", pidfile);
 	if (user) fprintf(fp, " -u '%s'", user);
@@ -1399,6 +1432,9 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 	int n, fl;
 	FILE *fp;
 
+	if (debuglevel) {
+		debug("do_cmd(%s, %p, %p)", b, output, op);
+	}
 	p = strchr(b, '\r');
 	if (p) *p = '\0';
 	p = strchr(b, '\n');
@@ -1642,7 +1678,7 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 		} else {
 			debug("write: no file");
 		}
-	} else if (debuglevel) {
+	} else {
 		debug("do_cmd: ignoring command starting with '%s'", p);
 	}
 }
@@ -1699,10 +1735,11 @@ static int server_by_weight(void)
 	int best_server = -1;
 	int best_load = -1;
 	int i, load;
+	int now = (int)time(NULL);
 
 	if (debuglevel) debug("server_by_weight()");
 	for (i = 0; i < nservers; i++) {
-		if (servers[i].status != 0 || servers[i].weight == 0) {
+		if (now-servers[i].status < blacklist_time || servers[i].weight == 0) {
 			continue;
 		}
 		load = servers[i].c/servers[i].weight;
@@ -1742,6 +1779,10 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 	pd = match_acl(client_acl, cli);
 	if (!pd) {
 		if (debuglevel) debug("add_client: denied by acl");
+		if (abuse_server != -1) {
+			upfd = try_server(abuse_server, 0, cli);
+			if (upfd != -1) goto Success;
+		}
 		goto Failure;
 	}
 
@@ -1752,7 +1793,7 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 		if (debuglevel) debug("lookup_client returns %d", clino);
 		if (clino != -1) {
 			index = clients[clino].cno;
-			if (index != emerg_server) {
+			if (index != emerg_server && index != abuse_server) {
 				upfd = try_server(index, 1, cli);
 				if (upfd != -1) goto Success;
 			}
@@ -2237,18 +2278,227 @@ static void mainloop_poll(void)
 }
 #endif	/* HAVE_POLL */
 
+#ifdef HAVE_KQUEUE
+
+static void dump_kevent(struct kevent *kev, int nfds)
+{
+	int i;
+	struct kevent *kptr;
+	for (i = 0; i < nfds; i++) {
+		kptr = &kev[i];
+		debug("kev[%d]: <i=%d, f=%d, f=%d, ff=%d, d=%d u=%p>", i,
+		    kptr->ident, kptr->filter, kptr->flags,
+		    kptr->fflags, kptr->data, kptr->udata);
+	}
+
+}
+
+static void mainloop_kqueue(void)
+{
+	int downfd, clilen;
+	int kq;
+	struct kevent *kev, *kptr;
+	connection *conn;
+	struct sockaddr_in cli_addr;
+	int i, j, nfds;
+	short downevents, upevents;
+
+	loopflag = 1;
+
+	kq = kqueue();
+	if (kq == -1)
+	{
+		perror("kqueue");
+		error("Error creating kernel queue");
+	}
+
+	kev = pen_malloc((connections_max*2+2)*sizeof *kev);
+
+	if (debuglevel) debug("mainloop_kqueue()");
+	while (loopflag) {
+		int n;
+
+		if (do_stats) {
+			if (webfile) webstats();
+			else textstats();
+			do_stats=0;
+		}
+		if (do_restart_log) {
+			if (logfp) {
+				fclose(logfp);
+				logfp = fopen(logfile, "a");
+				if (!logfp) 
+					error("Can't open logfile %s", logfile);
+			}
+			read_cfg(cfgfile);
+			do_restart_log=0;
+		}
+		kptr = kev;
+		nfds = 0;
+
+		/* FIXME: these two don't need to be done every loop;
+		 * they'll persist until the sockets are closed. But moving
+		 * them outside the loop adds complications (kptr initial
+		 * setting etc) */
+		EV_SET(kptr++, listenfd, EVFILT_READ, EV_ADD, 0, 0, 0);
+		nfds++;
+		if (ctrlfd != -1) {
+			EV_SET(kptr++, ctrlfd, EVFILT_READ, EV_ADD, 0, 0, 0);
+			nfds++;
+		}
+
+		/* add sockets from open connections */
+		if (debuglevel) debug("filling kqueue structure");
+		for (i = 0; i < connections_max; i++) {
+			if (conns[i].downfd == -1) continue;
+			upevents = downevents = 0;
+
+			if (conns[i].upn == 0)
+			{
+				EV_SET(kptr++, conns[i].downfd, EVFILT_READ,
+				    EV_ADD | EV_ONESHOT, 0, 0, &conns[i]);
+				nfds++;
+			}
+			else
+			{
+				EV_SET(kptr++, conns[i].upfd, EVFILT_WRITE,
+				    EV_ADD | EV_ONESHOT, 0, 0, &conns[i]);
+				nfds++;
+			}
+
+			if (conns[i].downn == 0)
+			{
+				EV_SET(kptr++, conns[i].upfd, EVFILT_READ,
+				    EV_ADD | EV_ONESHOT, 0, 0, &conns[i]);
+				nfds++;
+			}
+			else
+			{
+				EV_SET(kptr++, conns[i].downfd, EVFILT_WRITE,
+				    EV_ADD | EV_ONESHOT, 0, 0, &conns[i]);
+				nfds++;
+			}
+		}
+
+		if (debuglevel) dump_kevent(kev, nfds);
+
+		/* Wait for a connection from a client process. */
+		n = kevent(kq, kev, nfds, kev, nfds, NULL);
+		if (debuglevel) debug("n = %d", n);
+		if (n < 0 && errno != EINTR) {
+			perror("kevent");
+			error("Error on kevent");
+		}
+		if (n <= 0) continue;
+
+		if (debuglevel) debug("checking kqueue structure");
+		if (debuglevel) dump_kevent(kev, n);
+
+		for(j = 0 ; j < n ; j++)
+		{
+			kptr = &kev[j];
+
+			/* check listener */
+			if (kptr->ident == listenfd)
+			{
+				clilen = sizeof cli_addr;
+				downfd = accept(listenfd,
+					(struct sockaddr *) &cli_addr, &clilen);
+				if (downfd < 0) {
+					if (debuglevel) perror("accept");
+					continue;
+				}
+				if (clilen == 0) {
+					if (debuglevel) perror("clilen");
+					continue;
+				}
+				add_client(downfd, &cli_addr);
+				continue;
+			}
+
+			/* check control port */
+			if (ctrlfd != -1 && kptr->ident == ctrlfd)
+			{
+				clilen = sizeof cli_addr;
+				downfd = accept(ctrlfd,
+					(struct sockaddr *) &cli_addr, &clilen);
+				if (downfd < 0) {
+					if (debuglevel) perror("accept");
+					continue;
+				}
+				if (clilen == 0) {
+					if (debuglevel) perror("clilen");
+					continue;
+				}
+				do_ctrl(downfd, &cli_addr);
+				continue;
+			}
+
+			/* check sockets from open connections */
+			conn = (connection *)kptr->udata;
+			if (conn->downfd == -1) continue;
+
+			if (conn->downfd != kptr->ident)
+				downevents = 0;
+			else
+				downevents = kptr->filter;
+
+			if (conn->upfd != kptr->ident)
+				upevents = 0;
+			else
+				upevents = kptr->filter;
+
+			i = conn->i;
+			if (debuglevel) {
+				debug("conn = %d, upevents = %d, downevents = %d",
+					i, upevents, downevents);
+			}
+
+			if (downevents == EVFILT_READ) {
+				if (copy_up(i) < 0) {
+					close_conn(i);
+					continue;
+				}
+			}
+			if (upevents == EVFILT_READ) {
+				if (copy_down(i) < 0) {
+					close_conn(i);
+					continue;
+				}
+			}
+			if (downevents == EVFILT_WRITE) {
+				if (flush_down(i) < 0) {
+					close_conn(i);
+					continue;
+				}
+			}
+			if (upevents == EVFILT_WRITE) {
+				if (flush_up(i) < 0) {
+					close_conn(i);
+					continue;
+				}
+			}
+		}
+	}
+}
+#endif	/* HAVE_KQUEUE */
+
 static int options(int argc, char **argv)
 {
 	int c;
+	char b[1024];
 
 #ifdef HAVE_SSL
-	char *opt = "C:F:S:T:b:c:e:j:l:p:t:u:w:x:DHPWXadfhnrsE:K:G:A:ZRL:";
+	char *opt = "B:C:F:S:T:b:c:e:j:l:o:p:t:u:w:x:DHPQWXadfhnrsE:K:G:A:ZRL:";
 #else
-	char *opt = "C:F:S:T:b:c:e:j:l:p:t:u:w:x:DHPWXadfhnrs";
+	char *opt = "B:C:F:S:T:b:c:e:j:l:o:p:t:u:w:x:DHPQWXadfhnrs";
 #endif
 
 	while ((c = getopt(argc, argv, opt)) != -1) {
 		switch (c) {
+		case 'B':
+			a_server = optarg;
+			break;
 		case 'C':
 			ctrlport = optarg;
 			break;
@@ -2261,6 +2511,9 @@ static int options(int argc, char **argv)
 		case 'H':
 			http = 1;
 			break;
+		case 'Q':
+			use_kqueue = 1;
+			break;
 		case 'P':
 			use_poll = 1;
 			break;
@@ -2306,6 +2559,10 @@ static int options(int argc, char **argv)
 		case 'n':
 			nblock = 0;
 			break;
+		case 'o':
+			snprintf(b, sizeof b, "%s", optarg);
+			do_cmd(optarg, output_file, stdout);
+			break;
 		case 'p':
 			pidfile = optarg;
 			break;
@@ -2448,12 +2705,19 @@ int main(int argc, char **argv)
 		fclose(pidfp);
 	}
 
+#ifdef HAVE_KQUEUE
+	if (use_kqueue) mainloop_kqueue();
+#else
+	if (use_kqueue) error("You don't have kqueue()");
+#endif /* HAVE_KQUEUE */
+
 #ifdef HAVE_POLL
 	if (use_poll) mainloop_poll();
 #else
 	if (use_poll) error("You don't have poll()");
 #endif	/* HAVE_POLL */
-	if (!use_poll) mainloop_select();
+
+	if (!use_kqueue && !use_poll) mainloop_select();
 
 	if (debuglevel) debug("Exiting, cleaning up...");
 	if (logfp) fclose(logfp);
